# -*- coding: utf-8 -*-

from .serializable import Serializable
from .eq import DeepEqual, xhash
from ..utils.common import bstr

from collections import OrderedDict, UserDict
from pprint import pformat
import logging

logger = logging.getLogger(__name__)

# Depth of nesting of ODict.toString()
OD_toString_Nest = 0


class ODict(UserDict, Serializable, DeepEqual):
    """ Ordered dict that is not a subclass of dict and with a better __str__.
    """

    def __init__(self, *args, **kwds):
        """
        Parameters
        ----------

        Returns
        -------
        """

        super().__init__(*args, **kwds)
        self.__missing__ = None
        Serializable.__init__(self)

    # @property
    # def listoflists(self):
    #     return self.getListoflists()

    # @listoflists.setter
    # def listoflists(self, value):
    #     self.setListoflists(value)

    # def getListoflists(self):
    #     """ Returns a list of lists of key-value pairs where if the key is a tuple or frozenset, it is converted to a list.
    #     """
    #     ret = []
    #     for k, v in self.items():
    #         if issubclass(k.__class__, str):
    #             kk = k
    #         else:
    #             kk = list(k) if issubclass(k.__class__, (Collection)) else k
    #         ret.append([kk, v])
    #     return ret

    # def setListoflists(self, value):
    #     """ Sets the listoflists of this object. """
    #     def c2t(c):
    #         print(c)

    #         lst = [c2t(x) if issubclass(x.__class__, list) else x for x in c]
    #         print('== ', lst)
    #         return tuple(lst)
    #     d = dict(c2t(x) for x in value)
    #     self.clear()
    #     self.update(d)
    #     if 0:
    #         for item in value:
    #             kk = tuple(item[0])
    #             self[kk] = item[1]

    def toString(self, level=0,
                 tablefmt='rst', tablefmt1='simple', tablefmt2='simple',
                 matprint=None, trans=True, **kwds):
        """

        Parameters
        ----------

        Returns
        -------
        """
        global OD_toString_Nest

        # return 'OD' + str(type(self.data))+'*'+str(self.data)
        # return 'OD' + str(self.data)
        # return ydump(self.data)

        OD_toString_Nest += 1
        d = '<ODict '
        for n, v in self.data.items():
            #d += '    ' * OD_toString_Nest + '[ ' + str(n) + ' ]= '
            d += '"' + str(n) + '":' + '\n' if level == 0 else ' '
            s = bstr(v, level=level,
                     tablefmt=tablefmt, tablefmt1=tablefmt1, tablefmt2=tablefmt2,
                     matprint=matprint, trans=trans, **kwds)
            d = d + s
        OD_toString_Nest -= 1
        return d + '>'

    def get(self, name):
        """ Raise a ``KeyErrpr`` to change the default behavior of colections.Mapping to quietly return a None when a key is not found in the dict.
        """

        return self.data[name]
        res = super().__getitem__(name)
        if res is not None or name in self.data:
            return res
        logger.debug('%s is not found in %s.' % (name, self))
        raise KeyError()

    # def __repr__(self):
    #     """ returns string representation with details set according to debuglevel.
    #     """
    #     # return 'OD'+super().__repr__()
    #     level = int(logger.getEffectiveLevel()/10) - 1
    #     return self.toString(level=level)

    def __getstate__(self):
        """ Can be encoded with serializableEncoder
        Parameters
        ----------

        Returns
        -------

        """
        return OrderedDict(
            data=self.data,
            _STID=self._STID
        )

    def serializable(self):
        """ Can be encoded with serializableEncoder 
        Parameters
        ----------

        Returns
        -------

        """

        return OrderedDict(
            **self.data,
            _STID=self._STID
        )

    def hash(self):

        return xhash(hash_list=self.data.items())



from fdi.dataset.product import _Model_Spec as PPI
from .product import Product
from .numericparameter import NumericParameter
from .dateparameter import DateParameter
from .stringparameter import StringParameter
from .datatypes import Vector
from .dataset import CompositeDataset
from .tabledataset import TableDataset
from .arraydataset import ArrayDataset, Column
from ..pal.context import Context, MapContext
from .finetime import FineTime

import copy
from math import sin, cos
from os import path as op


class TP(Product):
    pass


class TC(Context):
    pass


class TM(MapContext):
    pass


# sub-classing testing class
# 'version' of subclass is int, not string

sp = copy.deepcopy(PPI)
sp['name'] = 'SP'
sp['metadata']['version']['data_type'] = 'integer'
sp['metadata']['version']['default'] = 9
sp['metadata']['type']['default'] = sp['name']
MdpInfo = sp['metadata']


class SP(Product):
    def __init__(self,
                 description='UNKNOWN',
                 typ_='SP',
                 creator='UNKNOWN',
                 version='9',
                 creationDate=FineTime(0),
                 rootCause='UNKNOWN',
                 startDate=FineTime(0),
                 endDate=FineTime(0),
                 instrument='UNKNOWN',
                 modelName='UNKNOWN',
                 mission='_AGS',
                 zInfo=None,
                 **kwds):
        metasToBeInstalled = copy.copy(locals())
        for x in ('self', '__class__', 'zInfo', 'kwds'):
            metasToBeInstalled.pop(x)

        self.zInfo = sp
        assert PPI['metadata']['version']['data_type'] == 'string'
        super().__init__(zInfo=zInfo, **metasToBeInstalled, **kwds)
        # super().installMetas(metasToBeInstalled)


def get_sample_product():
    """
    A complex product as a reference for testing and demo.

    ```
    prodx --+-- meta --+-- speed
            |
            +-- Temperature -+-- data=[768, ...] , unit=C
            |                |
            |                +-- meta --+-- T0
            |
            +-- results --+-- calibration -- data=[[109..]], unit=count
                          |
                          +-- Time_Energy_Pos --+-- Time   : data=[...]
                          |                     +-- Energy : data=[...]
                          |                     +-- y      : data=[...]
                          |                     +-- z      : data=[...]
                          |
                          +-- Image -- data = b'\87PNG', content='Content-type: image/png'
    ```

    """
    prodx = Product('A complex product')
    prodx.creator = 'Frankenstein'
    # add a parameter with validity descriptors to the product
    prodx.meta['speed'] = NumericParameter(
        description='an extra param',
        value=Vector((1.1, 2.2, 3.3)),
        valid={(1, 22): 'normal', (30, 33): 'fast'}, unit='meter')

    # an arraydsets
    a1 = [768, 767, 766, 4.4, 4.5, 4.6, 5.4E3]
    a2 = 'C'
    a3 = 'Temperature'
    a4 = ArrayDataset(data=a1, unit=a2, description='An Array')
    # metadata to the dataset
    a11 = 'T0'
    a12 = DateParameter('2020-02-02T20:20:20.0202',
                        description='meta of composite')
    # This is not the best as a4.T0 does not exist
    # a4.meta[a11] = a12
    # this does it a4.T0 = a12 or:
    setattr(a4, a11, a12)
    # put the arraydataset to the product with a name a3.
    prodx[a3] = a4

    compo = CompositeDataset()
    prodx['results'] = compo
    a5 = [[109, 289, 9], [88, 3455, 564]]
    a8 = ArrayDataset(data=a5, unit='count', description='array in composite')
    a10 = 'calibration'
    # put the dataset to the compositedataset. here set() api is used
    compo.set(a10, a8)
    # a tabledataset
    ELECTRON_VOLTS = 'eV'
    SECONDS = 'sec'
    METERS = 'm'
    t = [x * 1.0 for x in range(9)]
    e = [2 * x + 100 for x in t]
    y = [10 * sin(x*2*3.14/len(t)) for x in t]
    z = [10 * cos(x*2*3.14/len(t)) for x in t]
    x = TableDataset(description="A table")
    x["Time"] = Column(data=t, unit=SECONDS)
    x["Energy"] = Column(data=e, unit=ELECTRON_VOLTS)
    x["y"] = Column(data=y, unit=METERS)
    x["z"] = Column(data=z, unit=METERS)
    # set a tabledataset ans an arraydset, with a parameter in metadata
    compo['Time_Energy_Pos'] = x

    # an image
    fname = 'imageBlue.png'
    with open(op.abspath(op.dirname(__file__))+'/resources/'+fname, 'rb') as f:
        image = ArrayDataset(data=f.read(), description='An image in an array')
    image.file = fname
    image.content = 'Content-type: image/png'
    prodx['Image'] = image

    return prodx



# -*- coding: utf-8 -*-
from .odict import ODict
from .quantifiable import Quantifiable
from .eq import DeepEqual
from .copyable import Copyable
from .annotatable import Annotatable

from collections.abc import Container, Sized
import logging
# create logger
logger = logging.getLogger(__name__)
#logger.debug('level %d' %  (logger.getEffectiveLevel()))


class DataContainer(Annotatable, Copyable, DeepEqual, Container, Sized):
    """ A DataContainer is a composite of data and description.
    mh: note that There is no metadata.
    Implemented partly from AbstractDataWrapper.
    """

    def __init__(self, data=None, **kwds):
        """
        data: a Container. Default is None.

        Parameters
        ----------

        Returns
        -------

        """
        #print(__name__ + str(kwds))

        if data is None or issubclass(data.__class__, Container):
            self.setData(data)
        else:
            raise TypeError('DataContainer needs a Container to initialize, not ' +
                            type(data).__name__)
        super().__init__(**kwds)

    @property
    def data(self):
        """
        Parameters
        ----------

        Returns
        -------

        """ 
        return self.getData()

    @data.setter
    def data(self, newData):
        """ Property of the data of this datawrapper object.

        Parameters
        ----------

        Returns
        -------

        """
        self.setData(newData)

    def setData(self, data):
        """ Replaces the current DataData with specified argument. 
        mh: subclasses can override this to add listener whenevery data is
        replaced.

        Parameters
        ----------

        Returns
        -------

        """
        self._data = data

    def getData(self):
        """ Returns the data in this dw

        Parameters
        ----------

        Returns
        -------

        """
        try:
            return self._data
        except AttributeError:
            od = ODict()
            self._data = od
            return od

    def hasData(self):
        """ Returns whether this data wrapper has data. 


        Parameters
        ----------

        Returns
        -------

        """

        return self.getData() is not None and len(self.getData()) > 0

    def __contains__(self, x):
        """
        """
        try:
            return x in self._data
        except AttributeError:
            return False

    def __len__(self, *args, **kwargs):
        """ size of data
        """
        try:
            return self._data.__len__(*args, **kwargs)
        except AttributeError:
            return 0


class DataWrapper(DataContainer, Quantifiable):
    """ A DataWrapper is a composite of data, unit and description.
    mh: note that all data are in the same unit. There is no metadata.
    Implemented from AbstractDataWrapper.
    """

    def __init__(self, *args, **kwds):
        """ 
        """
        super().__init__(*args, **kwds)


class DataWrapperMapper():
    """ Object holding a map of data wrappers. """

    def __init__(self, *args, **kwds):
        """ 
        """
        super().__init__(*args, **kwds)

    def getDataWrappers(self):
        """ Gives the data wrappers, mapped by name.

        Parameters
        ----------

        Returns
        -------

        """
        return self._sets
